import re

WORD_START = re.compile(r"(?<!^)(?=[A-Z])")


def to_snakecase(string: str):
    return WORD_START.sub("_", string).lower()


def snakecase_to_capwords(snakecase: str) -> str:
    return "".join(word.capitalize() for word in snakecase.split("_"))


def speech_duration(
    text: str,
    words_per_second: float = 3.3,
    min_duration: float = 1.0
) -> float:
    return max(len(text.split()) / words_per_second, min_duration)




import inspect
import json
from collections import UserDict
from collections.abc import Callable
from collections.abc import Iterator
from functools import cache
from pathlib import Path
from typing import Any

import pygame

type Asset = Any  # type: ignore


def load_font(path: Path):
    return cache(lambda size=30: pygame.font.Font(path, size))


def load_json(path: Path):
    return json.loads(path.read_text())


def load_yaml(path: Path):
    import yaml

    return yaml.safe_load(path.read_text())


LOADERS: dict[str, Callable] = {
    **{_: pygame.image.load for _ in {".png", ".jpg", ".jpeg", ".gif"}},
    **{_: pygame.mixer.Sound for _ in {".wav", ".mp3", ".ogg"}},
    **{_: load_font for _ in {".ttf"}},
    **{_: load_yaml for _ in {".yaml", ".yml"}},
    **{_: load_json for _ in {".json"}},
}


def get_assets_path_from_caller_frame() -> Path:
    try:
        return Path(inspect.stack()[2].filename).parent / "assets"
    except IndexError:
        return Path("assets")


def find_assets_by_name(path: Path, name: str) -> Iterator[Path]:
    yield from sorted(
        (m for m in path.glob(f"{name}*.*") if m.suffix in LOADERS),
        # shortest match first
        key=lambda _: len(str(_)),
    )


def get_asset(path: Path, name: str) -> Asset:
    match next(find_assets_by_name(path.parent, name), None):
        case Path() as p if p.is_dir():
            asset = Assets(p)
        case Path() as p:
            asset = LOADERS[p.suffix](p)
        case None:
            raise LookupError(f"Asset not found: {path / name}")
    return asset


class Assets(UserDict):
    """
    Provides attribute access to static assets in a directory.

    >>> import tempfile
    >>> tempdir = tempfile.TemporaryDirectory()
    >>> Path(f"{tempdir.name}/foo.gif").write_bytes(
    ...     bytes.fromhex('4749463839610100010000ff002c00000000010001000002003b'))
    26
    >>> _ = pygame.init()
    >>> _ = pygame.display.set_mode((1, 1), pygame.HIDDEN)
    >>> Assets(tempdir.name).foo.get_size()
    (1, 1)
    """

    def __init__(self, path: str | Path | None = None) -> None:
        super().__init__()
        match path:
            case Path():
                ...
            case str(s):
                path = Path(s)
            case None:
                path = get_assets_path_from_caller_frame()
            case _:
                raise TypeError(f"Expected Path or str, got {type(path)}")
        if not path.is_dir():
            raise ValueError(f"Path does not exist: {path}")
        self.__dict__["path"] = path

    def __getitem__(self, name: str) -> Asset:
        if not (asset := self.data.get(name)):
            try:
                asset = self.data[name] = get_asset(self.__dict__["path"] / name, name)
            except LookupError as e:
                raise KeyError(name) from e
        return asset

    def __getattr__(self, name: str) -> Asset:
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError from e

    def load_all(self) -> None:
        for child in list(self.path.iterdir()):
            self[child.stem]



import binascii
import struct
from base64 import b64decode
from base64 import b64encode


def rhash(n: int) -> str:
    """Reversibly hash a number to a 6 character string."""
    return b64encode(struct.pack("<q", n * 387420469 % 1000000000)).decode("ascii")[:6]


def unrhash(h: str) -> int:
    """Unhash a 6 character string to a number."""
    try:
        return struct.unpack("<q", b64decode(h + "AAAAA="))[0] * 703225629 % 1000000000
    except (struct.error, binascii.Error) as e:
        raise ValueError("Invalid hash") from e




"""A simple animation system."""

from bisect import bisect
from collections.abc import Callable
from collections.abc import Iterator
from typing import Generic
from typing import Protocol
from typing import TypeVar
from typing import cast
from typing import runtime_checkable

import pygame

T = TypeVar("T")


EasingFn = Callable[[float], float]
LerpFn = Callable[[T, T, float], T]


@runtime_checkable
class Lerpable(Generic[T], Protocol):
    """Can be plugged into standard linear interpolation algorithm."""

    def __add__(self: T, other: T) -> T:
        ...

    def __mul__(self: T, other: float) -> T:
        ...

    def __sub__(self: T, other: T) -> T:
        ...


def lerp(start, end, quotient):
    """Linear interpolation between two values."""
    if isinstance(start, dict) and isinstance(end, dict):
        return {
            key: lerp(value, end[key], quotient)
            for key, value in start.items()
        }

    if hasattr(start, "lerp") and callable(start.lerp):
        return start.lerp(end, quotient)

    if isinstance(start, Lerpable):
        return start + ((end - start) * quotient)

    # can't interpolate
    return start


def animate(
    keyframes: dict[float, T | tuple[T, EasingFn]] | list[T],
    get_quotient: Callable[[], float],
    lerp: LerpFn = lerp,
) -> Iterator[T]:
    """
    Get the frame of animation that interpolates between keyframes.

    >>> frames = {
    ...     0.0: 2.0,
    ...     0.5: 8.0,
    ...     1.0: 3.0,
    ... }
    >>> from pygskin import Timer  # doctest: +ELLIPSIS
    >>> timer = Timer(3000)
    >>> anim = animate(frames, timer.quotient)
    >>> next(anim)
    2.0
    >>> timer.tick(1500)
    >>> next(anim)
    8.0
    >>> timer.tick(750)
    >>> next(anim)
    5.5
    >>> timer.tick(750)
    >>> next(anim)
    3.0
    """
    if isinstance(keyframes, list):
        num_frames = len(keyframes)
        if num_frames < 2:
            raise ValueError("At least two frames are required")
        keyframes = {i / (num_frames - 1): keyframes[i] for i in range(num_frames)}

    keys = sorted(keyframes.keys())
    if min(keys) < 0.0 or max(keys) > 1.0:
        raise ValueError("Keyframes must span [0, 1]")

    while True:
        easing = None
        try:
            key = pygame.math.clamp(get_quotient(), 0.0, 1.0)
        except StopIteration:
            return
        key_pos = bisect(keys, key)

        start_key = keys[max(0, key_pos - 1)]
        if isinstance(start_frame := keyframes[start_key], tuple):
            start_frame, easing = start_frame

        if start_key == keys[-1]:
            yield cast(T, start_frame)
            continue

        end_key = keys[min(len(keys) - 1, key_pos)]
        if isinstance(end_frame := keyframes[end_key], tuple):
            end_frame, _ = end_frame

        if end_key == keys[0]:
            yield cast(T, end_frame)
            continue

        duration = end_key - start_key
        quotient = (key - start_key) / duration

        if easing:
            quotient = easing(quotient)

        yield lerp(start_frame, end_frame, quotient)

